#include <string>
#include <iomanip>
#include "TSeries.hh"
#include "Time.hh"
#include "Interval.hh"
#include "Histogram1.hh"
#include "Complex.hh"
#include "IIRFilter.hh"
#include "constant.hh"

// IIR-filtering kOffsetFactor seconds of time series
// 1.0 to 2.0 second
#define kOffsetFactor 3/2

/// applies a filter to data
// extern void ApplyFilter(int,const int, const float*,float*,
//  			const double*,const double*, int);
/// gets filter coefficients from data
// extern void GetFilterCoef(const int,const int,const float,const float, 
//  			  double*,double*,const int);

/** ChannelStatus: 
  * @memo Class to hold channel information
  * @author Rauha Rahkola, Robert Schofield
  * @version ver. 0.1, Wednesday, February 27th, 2002
  *********************************************************************/
class ChannelStatus
{
  
public:
  /*
   * Channel Modes
   */
  enum { // Event status-- not event, during event, waiting for next event
    /// Normal-- not currently in event
    kNone = 0,     // normal
    /// Currently in an event
    kEvent= 1,     // during event
    /// Waiting for the event to continue/finish
    kWait = 2,     // waiting event end
    /// Event has finished
    kDone = 3      // finished with event
  } kStatus;

  /*
   * Channel identifiers/parameters
   */
  /// channel name
  string fName;
  /// sampling frequency
  int fSampleFreq;
  /// dead time to wait for another glitch
  double fDeadTime;
  /// thresholds: min (event above this level always triggered), max
  float fThreshold;
  /// whether to use given Operating State Condition
  bool fUseOSC;
  /// trigger comment
  string fTrigComment;
  /// debug level
  int  fDebug;

  /* 
   * Channel filter parameters
   */
  /// IIR Filter
  IIRFilter *fFilter;
  /// poles and zeroes of IIR Filter
  std::vector<dComplex> fPoles, fZeroes;
  /// gain of IIR Filter
  double fGain;
  /// end time of the previously filtered time series
  Time fPrevTime;
  /// cut off frequencies (lower and higher edges)
  //float fCutOffFreqLow,fCutOffFreqHi;
  /// filter order & mode (0:lowpass 1:highpass 2:bandpass)
  //int fFilterOrder, fFilterMode;
  /// IIR filter coefficients (denominator & numerator) 
  //double *fCoefA,*fCoefB;

  /* 
   * Channel data
   */
  /// Pointers to hold current and previous second of raw data.
  const TSeries *fTSnew;
  /// Histogram data
  Histogram1 fHist; // updates every log interval
  Histogram1 fHistAll;
  Histogram1 fTimeHist; // updates every log interval with duration of glitch
  Histogram1 fTimeHistAll;
  Histogram1 fShortHist; // updates every log interval with single-bin glitches

  /* 
   * Channel event/log statistics
   */
  typedef struct {
    Time         startTime; // start time of event
    Time         endTime;   // end time of event
    double       amplitude; // abs(max amplitude during event)
    double       variance;  // variance during event--> rms (when finished)
    long         statCount; // number of datapts in event
    double       waitVar;   // variance during waiting period
    long         waitCount; // number of datapts in waiting period
    unsigned int status;    // status of event (0=none; 1=in event; 2=waiting)
  } absEvent;
  /// holds the event statistics for each threshold
  std::vector<absEvent> fEventStatus;

  /* 
   * Channel (filtered) data statistics
   */
  /// Mean, Var, & Kurt (not during event) :NOTE:RMS of unfiltered data is SDev
  float fMean, fVar, fRMS;
  double fQuart, fKurt;
  /// number of statistics calculated
  int fStatCount;

  /* 
   * Standard class functions
   */
  /// Create a channel info structure
  ChannelStatus(int debug=0) : fDebug(debug), fGain(1) {
    fTSnew = 0;
    Reset( TSeries() );
  }
  
  /// Create a channel info structure
  ChannelStatus(const ChannelStatus& ch) {
    fTSnew = 0;
    *this = ch; 
  }
  
  /// Destroy a channel info structure
  ~ChannelStatus(void) {
    if (fTSnew) delete fTSnew;
    fTSnew = 0;
  }
  
  /// copy a channel info structure
  ChannelStatus& operator = (const ChannelStatus& ch) {
    Reset( *ch.fTSnew );
    fName            = ch.fName;
    fDeadTime        = ch.fDeadTime;
    fUseOSC          = ch.fUseOSC;
    fThreshold       = ch.fThreshold;
    fEventStatus     = ch.fEventStatus;
    fMean            = ch.fMean;
    fVar             = ch.fVar;
    fRMS             = ch.fRMS;
    fQuart           = ch.fQuart;
    fKurt            = ch.fKurt;
    fDebug           = ch.fDebug;
    fStatCount       = ch.fStatCount;
    fTrigComment     = ch.fTrigComment;

    ResetFilter( ch.fPoles, ch.fZeroes, ch.fSampleFreq, ch.fGain );

    return *this;
  }  // ChannelStatus& operator = (const ChannelStatus &ch) ...

  /*
   * Other class functions
   */
  /// Set up the filter coefficients, based on two C-strings
  void SetupCoefs( const char *pstr, const char *zstr, 
		   const char *kstr="", const char *units="Hz", const int debug=0 ) {
    const char* ptr = pstr;
    if (debug > 2) cout << "Poles string length: "<< strlen(ptr)<<endl;
    if ( strlen(ptr) <= 2 ) {
      fPoles.clear();
      fZeroes.clear();
      return;
    }
    double twopi= 2 * 3.1415926535;
    if ( !strcmp(units,"Hz") ) {
	twopi = 1;
    } else if ( !strcmp(units,"rad/s") ) {
        twopi = 2 * 3.1415926535;
    }
    while (*ptr) {
      double re = strtod(ptr, const_cast<char**>(&ptr));
      double im = strtod(ptr, const_cast<char**>(&ptr));
      fPoles.push_back( dComplex(re,im) / twopi );
      if ( debug ) cout << "pole " << fPoles.size() << " : "
			<< fPoles.back().Real() << " + "
			<< fPoles.back().Imag() << 'i'<< endl;
    }

    ptr = zstr;
    while (*ptr) {
      double re = strtod(ptr, const_cast<char**>(&ptr));
      double im = strtod(ptr, const_cast<char**>(&ptr));
      fZeroes.push_back(dComplex(re,im)/twopi);
      if ( debug ) cout << "zero " << fZeroes.size() << " : "
			<< re << " + " << im << "i" << endl;
    }

    ptr = kstr;
    if (strlen(ptr)) fGain = strtod(ptr, NULL);
  }

  /// clear all variables
  void Reset( const TSeries &ts ) {
    if (fTSnew) delete fTSnew;
    if ( ! ts.isEmpty() ) fTSnew = new const TSeries( ts );
    else fTSnew = new const TSeries();

    fName                            = "";
    fTrigComment                     = "";
    fUseOSC                          = false;
    fSampleFreq                      = 0;
    fGain                            = 1;
    fMean = fVar = fRMS              = 0;
    fQuart = fKurt                   = 0;
    fDeadTime = fThreshold           = 0;
    for (unsigned int i=0; i<fEventStatus.size(); i++) {
      fEventStatus[i].startTime = fEventStatus[i].endTime   = Time(0);
      fEventStatus[i].amplitude = fEventStatus[i].variance  = 0.0;
      fEventStatus[i].waitVar   = 0.0;
      fEventStatus[i].statCount = fEventStatus[i].waitCount = 0;
      fEventStatus[i].status    = kNone;
    }
  }

  /// reset IIR Filter
  void ResetFilter( const std::vector<dComplex>& poles, 
		    const std::vector<dComplex>& zeroes,
		    const int sampleFreq, const double gain ) {
    fPoles    = poles;
    fZeroes   = zeroes;
    fSampleFreq = sampleFreq;
    fGain     = gain;

    delete fFilter;
    fFilter = new IIRFilter( fPoles.size(), &fPoles[0], 
			     fZeroes.size(), &fZeroes[0], sampleFreq, fGain );
  }
  
  /// reset histograms
  void ResetHistograms(int nBins, double min, double max) {
    string name;
    if (fTrigComment == "") name = fName;
    else name = fTrigComment;
    fHist = Histogram1(name.c_str(),nBins,min,max,"Glitch size (counts)","");
    name += "_ALL";
    fHistAll = Histogram1(name.c_str(),nBins,min,max,
			  "Glitch size (counts)","");
    name = fName + "_Time";
    fTimeHist = Histogram1(name.c_str(),nBins,min,max,
			   "Glitch size (counts)","Dead time (s)");
    name += "_ALL";
    fTimeHistAll = Histogram1(name.c_str(),nBins,min,max,
			      "Glitch size (counts)","Dead time (s)");
    name = fName + "_SingleBin";
    fShortHist = Histogram1(name.c_str(),nBins,min,max,
			    "Glitch size (counts)","");
    fEventStatus.resize(nBins + 2);
    for( unsigned int i=0; i < fEventStatus.size(); i++ ) {
      fEventStatus[i].startTime =
	fEventStatus[i].endTime = fTSnew->getStartTime();
      fEventStatus[i].amplitude = fEventStatus[i].variance =
	fEventStatus[i].waitVar = 0.0;
      fEventStatus[i].statCount = fEventStatus[i].waitCount = 0;
      fEventStatus[i].status    = kNone;
    }
  }
  
  /// clear histogram (f(Time)Hist only; f(Time)HistAll is continuous)
  void ClearHistogram(void) {
    int nbin = fHist.GetNBins();
    double* zero = new double[nbin+2];
    memset(zero,0,(nbin+2) * sizeof(double));
    fHist.SetBinContents(zero);
    fHist.SetNEntries(0);
    fHist.PutStats(zero);
    fTimeHist.SetBinContents(zero);
    fTimeHist.SetNEntries(0);
    fTimeHist.PutStats(zero);
    fShortHist.SetBinContents(zero);
    fShortHist.SetNEntries(0);
    fShortHist.PutStats(zero);
    delete[] zero;
  }
  
  /// find the length of an event
  double GetEventDuration(int idx=1) const {
    return fEventStatus[idx].endTime - fEventStatus[idx].startTime;
  }
  
  /// Is the new channel data a continuation of the previous data
  bool IsTSContinuous(void) const {
    if ( Almost( fPrevTime, fTSnew->getStartTime(), 5 ) ) return true;
    else if ( fTSnew->isEmpty() ) return true;
//      else return false;
    return true;
  }

  /// set the debug level
  void setDebug(int debug) {
    fDebug = debug;
  }

  /// calculate the rms of the filtered data
  float CalcVar(const TSeries* ts) const {
    /* Procedure: Check for data.  If so, iterate through the datapts and 
     *  calculate the variance.  (Later we will use the variance to find the
     *  RMS)
     */
    if( ts->isEmpty() ) return 0;

    const double *data = (const double *) ts->refData();
    double var = 0;
    //    int step = fSampleFreq / 256;
    //    if ( !step ) return false;
    
    for ( int i = 0; i < fSampleFreq; i++      )
      var += data[i] * data[i];
    
    return ( var/fSampleFreq + fVar * fStatCount ) / (fStatCount + 1);
  }

  /// calculate the kurtosis of the filtered data
  double CalcQuart(const TSeries* ts) const {
    if ( ts->isEmpty() ) return 0;

    const double *data = (const double *) ts->refData();
    double kurt = 0;
    //    int step = fSampleFreq / 256;
    //    if ( !step ) return false;

    for ( int i = 0; i < fSampleFreq; i++      ) {
      kurt += data[i] * data[i] * data[i] * data[i];
    }

    return (kurt/fSampleFreq + fQuart * fStatCount)/(fStatCount + 1);
  }

  /// Update the array with new filtered data
  void UpdateStats() {
    // Decided against checking for events, as this doesn't actually mean
    // all events are excluded from the statistics-- only those which span
    // the TimeStep boundaries.
    // Note: All calculated stats assume 0 mean, which works for >1Hz filtering
    //    if ( (fEventStatus == kNone) ) {
      fVar   = CalcVar( fTSnew );
      fRMS   = sqrt( fVar );
      fQuart = CalcQuart( fTSnew );
      if (fVar) fKurt  = fQuart / (fVar * fVar) - 3;
      if (fDebug > 2) cout <<"fVar = "<<fVar <<"fRMS = "<<fRMS
			   <<"fQuart = "<<fQuart <<"fKurt = "<<fKurt <<endl;
      fStatCount++;
    //    }
  }

  // glitch status check
  /// Dump the poles of the IIR Filter
  std::ostream& DumpPoles( std::ostream &os ) const {
    if (fDebug > 3) fFilter->dumpSPlaneRoots(os);
    os << '"';
    for( unsigned int i=0; i < fPoles.size(); i++ ) {
      if (i) os <<' ';
      os << fPoles[i].Real() <<' ' <<fPoles[i].Imag();
    }
    return ( os << "\" " );
  }

  /// Dump the zeroes of the IIR Filter
  std::ostream& DumpZeroes( std::ostream &os ) const {
    os << '"';
    for( unsigned int i=0; i < fZeroes.size(); i++ ) {
      if (i) os <<' ';
      os << fZeroes[i].Real() <<' ' <<fZeroes[i].Imag();
    }
    return ( os << "\" " );
  }

  /// Filter the data
  void FilterData( const TSeries &daccTS ) {
    fPrevTime = fTSnew->getEndTime();
    delete fTSnew;
    if ( fPoles.size() == 0 ) fTSnew = new const TSeries(daccTS);
    else // apply IIR filter
      fTSnew = new TSeries( fFilter->apply(daccTS) );
  }

  /** Handle an event: place event in the proper bin
    * return: true if one or more events has finished
    */
  bool HandleEvent( const double& data, const Time& bintime,
		    const bool flush=false ) {
    bool finished(false);
    for (unsigned int i=1; i < fEventStatus.size(); i++) {
      //=====================  Data above threshold
      if ( abs(data) > fHist.GetBinLowEdge(i) ) {
	//--------- Special case: 0 dead time to histogram datapts
	if ( !(fDeadTime) ) {
	  fEventStatus[i].amplitude = abs(data);
	  fEventStatus[i].startTime = bintime;
	  fEventStatus[i].endTime = bintime + double(1)/(fSampleFreq);
	  fEventStatus[i].status  = kDone;
	  finished |= true;
	}
	//--------- Case 1: Start of event
	if (fEventStatus[i].status == kNone) {
	  fEventStatus[i].startTime = bintime;
	  fEventStatus[i].endTime   = bintime;
	  fEventStatus[i].amplitude = abs(data);
	  fEventStatus[i].variance  = data * data;
	  fEventStatus[i].statCount = 1;
	  fEventStatus[i].waitVar   = 0.0;
	  fEventStatus[i].waitCount = 0;
	  fEventStatus[i].status    = kEvent;
	  if( fDebug > 2 ) cout <<fHist.GetBinLowEdge(i)
				<<": Start   datapt-- GPS: "
				<<setprecision(15) <<bintime.totalS()
				<<"  \tvalue: " <<data <<endl;
	  finished |= false;
	//--------- Case 2: Event continues
	} else {
	  fEventStatus[i].endTime   = bintime;
	  if ( abs(data) > fEventStatus[i].amplitude ) 
	    fEventStatus[i].amplitude = abs(data);
	  fEventStatus[i].variance   += data * data + fEventStatus[i].waitVar;
	  fEventStatus[i].statCount  += 1 + fEventStatus[i].waitCount;
	  fEventStatus[i].waitVar     = 0;
	  fEventStatus[i].waitCount   = 0;
	  fEventStatus[i].status      = kEvent;
	  if( fDebug > 2 ) cout <<fHist.GetBinLowEdge(i)
				<<": Continu datapt-- GPS: "
				<<setprecision(15) <<bintime.totalS()
				<<"  \tvalue: "<<data <<endl;
	  finished |= false;
	}

	//=====================  Data below threshold    
      } else { 
	//--------- Case 1: Still in Event, or still Waiting
	if ( fEventStatus[i].status != kNone &&
	     (fDeadTime >= (bintime - fEventStatus[i].endTime)) ) {
	  fEventStatus[i].waitVar  += data * data;
	  fEventStatus[i].waitCount++;
	  fEventStatus[i].status    = kWait;
	  if( fDebug > 2 ) cout <<fHist.GetBinLowEdge(i)
				<<": Waiting datapt-- GPS: "
				<<setprecision(15) <<bintime.totalS()
				<<"  \tvalue: "<<data <<endl;
	  finished |= false;
	//--------- Case 2: Tired of waiting-- end of event
	} else if ( fEventStatus[i].status != kNone && 
		    (fDeadTime < (bintime - fEventStatus[i].endTime)) ) {
	  fEventStatus[i].endTime += Interval(1.0/fSampleFreq);
	  fEventStatus[i].variance = sqrt( fEventStatus[i].variance / 
					   fEventStatus[i].statCount );
	  fEventStatus[i].statCount= fEventStatus[i].waitCount = 0;
	  fEventStatus[i].waitVar  = 0.0;
	  fEventStatus[i].status   = kDone;
	  if( fDebug > 2 ) cout <<fHist.GetBinLowEdge(i)
				<<": Finish  datapt-- GPS: "
				<<setprecision(15) <<bintime.totalS()
				<<"  \tvalue: "<<data <<endl;
	  finished |= true;
	//--------- Case 3: Not in event in the first place
	} else finished |= false;
      } // end if ( abs(data) > fHist.GetBinLowEdge(i) ) ...
      if (flush && (fEventStatus[i].status != kNone)) {
	fEventStatus[i].status = kDone;
	finished |= true;
      }

    } // end for (int i=1; i < fHist.GetNBins(); i++) ...
    return finished;
  } // end bool HandleEvent( const double& data, const Time& bintime,...

}; // end of ChannelStatus class definition
// end of ChannelStatus class section
